@@ -10,6 +10,7 @@ use strict;
 use integer;
 
 use vars qw( $ERROR );
+use MARC::File::Encode qw( marc_to_utf8 );
 
 use MARC::File;
 use vars qw( @ISA ); @ISA = qw( MARC::File );
@@ -51,19 +52,7 @@ sub _next {
     my $usmarc = <$fh>;
 
     # remove illegal garbage that sometimes occurs between records
-    $usmarc =~ s/^[ \x00\x0a\x0d]+//;
-
-    return unless $usmarc;
-
-    if ( length($usmarc) < 5 ) {
-        $self->_warn( "Couldn't find record length" );
-    }
-
-    $reclen = substr($usmarc,0,5);
-
-    if ( $reclen !~ /^\d{5}$/ or $reclen != length($usmarc) ) {
-        $self->_warn( "Invalid record length \"$reclen\"" );
-    }
+    $usmarc =~ s/^[ \x00\x0a\x0d\x1a]+//;
 
     return $usmarc;
 }
@@ -123,9 +112,9 @@ sub decode {
         $location = 'in record 1';
         $text = $self=~/MARC::File/ ? shift : $self;
     }
-
     my $filter_func = shift;
 
+    # ok this the empty shell we will fill
     my $marc = MARC::Record->new();
 
     # Check for an all-numeric record length
@@ -133,8 +122,9 @@ sub decode {
         or return $marc->_warn( "Record length \"", substr( $text, 0, 5 ), "\" is not numeric $location" );
 
     my $reclen = $1;
-    ($reclen == length($text))
-        or $marc->_warn( "Invalid record length $location: Leader says $reclen bytes, but it's actually ", length( $text ) );
+    my $realLength = bytes::length( $text );
+    $marc->_warn( "Invalid record length $location: Leader says $reclen " . 
+        "bytes but it's actually $realLength" ) unless $reclen == $realLength;
 
     (substr($text, -1, 1) eq END_OF_RECORD)
         or $marc->_warn( "Invalid record terminator $location" );
@@ -142,8 +132,7 @@ sub decode {
     $marc->leader( substr( $text, 0, LEADER_LEN ) );
 
     # bytes 12 - 16 of leader give offset to the body of the record
-    my $data_start = 0 + substr( $text, 12, 5 );
-
+    my $data_start = 0 + bytes::substr( $text, 12, 5 );
 
     # immediately after the leader comes the directory (no separator)
     my $dir = substr( $text, LEADER_LEN, $data_start - LEADER_LEN - 1 );  # -1 to allow for \x1e at end of directory
@@ -175,10 +164,15 @@ sub decode {
         ($offset + $len <= $reclen)
             or $marc->_warn( "Directory entry $location runs off the end of the record tag $tagno" );
 
-        my $tagdata = substr( $text, $data_start + $offset, $len );
+        my $tagdata = bytes::substr( $text, $data_start+$offset, $len ); 
 
-        ($len == length($tagdata))
-            or $marc->_warn( "Invalid length in directory for tag $tagno $location" );
+        # if utf8 the we encode the string as utf8
+        if ( $marc->encoding() eq 'UTF-8' ) {
+            $tagdata = marc_to_utf8( $tagdata );
+        }
+
+        $marc->_warn( "Invalid length in directory for tag $tagno $location" )
+            unless ( $len == bytes::length($tagdata) );
 
         if ( substr($tagdata, -1, 1) eq END_OF_FIELD ) {
             # get rid of the end-of-tag character
@@ -190,12 +184,10 @@ sub decode {
 
         warn "Specs: ", join( "|", $tagno, $len, $offset, $tagdata ), "\n" if $MARC::Record::DEBUG;
 
-
         if ( $filter_func ) {
             next unless $filter_func->( $tagno, $tagdata );
         }
 
-
         if ( ($tagno =~ /^\d+$/) && ($tagno < 10) ) {
             $marc->append_fields( MARC::Field->new( $tagno, $tagdata ) );
         } else {
@@ -281,7 +273,8 @@ sub _build_tag_directory {
                 push( @fields, $str );
 
                 # Create directory entry
-                my $len = length $str;
+                my $len = bytes::length( $str );
+
                 my $direntry = sprintf( "%03s%04d%05d", $field->tag, $len, $dataend );
                 push( @directory, $direntry );
                 $dataend += $len;
@@ -321,7 +314,6 @@ sub encode() {
     # Glomp it all together
     return join("",$marc->leader, @$directory, END_OF_FIELD, @$fields, END_OF_RECORD);
 }
-
 1;
 
 __END__
